@@ -13,6 +13,8 @@ class Accessor;
 
 class ExtensionAPI {
 public:
+	virtual ~ExtensionAPI() {
+	}
 	enum Pane { paneEditor=1, paneOutput=2, paneFindOutput=3 };
 	virtual sptr_t Send(Pane p, unsigned int msg, uptr_t wParam=0, sptr_t lParam=0)=0;
 	virtual char *Range(Pane p, int start, int end)=0;
@@ -21,6 +23,7 @@ public:
 	virtual void Trace(const char *s)=0;
 	virtual char *Property(const char *key)=0;
 	virtual void SetProperty(const char *key, const char *val)=0;
+	virtual void UnsetProperty(const char *key)=0;
 	virtual uptr_t GetInstance()=0;
 	virtual void ShutDown()=0;
 	virtual void Perform(const char *actions)=0;
@@ -39,6 +42,10 @@ public:
 	virtual bool Clear()=0;
 	virtual bool Load(const char *filename)=0;
 
+	virtual bool InitBuffer(int) { return false; }
+	virtual bool ActivateBuffer(int) { return false; }
+	virtual bool RemoveBuffer(int) { return false; }
+
 	virtual bool OnOpen(const char *) { return false; }
 	virtual bool OnSwitchFile(const char *) { return false; }
 	virtual bool OnBeforeSave(const char *) { return false; }
@@ -54,6 +61,7 @@ public:
 	virtual bool OnUpdateUI() { return false; }
 	virtual bool OnMarginClick() { return false; }
 	virtual bool OnMacro(const char *, const char *) { return false; }
+	virtual bool OnUserListSelection(int, const char *) { return false; }
 
 	virtual bool SendProperty(const char *) { return false; }
 };
