@@ -331,11 +331,15 @@ static int cf_pane_textrange(lua_State *L) {
 		int cpMin = static_cast<int>(luaL_checknumber(L, 2));
 		int cpMax = static_cast<int>(luaL_checknumber(L, 3));
 
-		char *range = host->Range(p, cpMin, cpMax);
-		if (range) {
-			lua_pushstring(L, range);
-			delete []range;
-			return 1;
+		if (cpMax >= 0) {
+			char *range = host->Range(p, cpMin, cpMax);
+			if (range) {
+				lua_pushstring(L, range);
+				delete []range;
+				return 1;
+			}
+		} else {
+			raise_error(L, "Invalid argument 2 for <pane>:textrange.  Positive number or zero expected.");
 		}
 	} else {
 		raise_error(L, "Not enough arguments for <pane>:textrange");
@@ -400,7 +404,7 @@ static int cf_pane_findtext(lua_State *L) {
 		}
 
 		if (!hasError) {
-			int result = host->Send(p, SCI_FINDTEXT, reinterpret_cast<uptr_t&>(flags), reinterpret_cast<sptr_t>(&ft));
+			int result = host->Send(p, SCI_FINDTEXT, static_cast<uptr_t>(flags), reinterpret_cast<sptr_t>(&ft));
 			if (result >= 0) {
 				lua_pushnumber(L, ft.chrgText.cpMin);
 				lua_pushnumber(L, ft.chrgText.cpMax);
@@ -600,7 +604,7 @@ static int cf_pane_match_generator(lua_State *L) {
 	ft.lpstrText = const_cast<char*>(text);
 
 	if (ft.chrg.cpMax > ft.chrg.cpMin) {
-		int result = host->Send(pmo->pane, SCI_FINDTEXT, reinterpret_cast<uptr_t&>(pmo->flags), reinterpret_cast<sptr_t>(&ft));
+		int result = host->Send(pmo->pane, SCI_FINDTEXT, static_cast<uptr_t>(pmo->flags), reinterpret_cast<sptr_t>(&ft));
 		if (result >= 0) {
 			pmo->startPos = ft.chrgText.cpMin;
 			pmo->endPos = pmo->endPosOrig = ft.chrgText.cpMax;
@@ -1709,6 +1713,10 @@ bool LuaExtension::OnDoubleClick() {
 	return CallNamedFunction("OnDoubleClick");
 }
 
+bool LuaExtension::OnUpdateUI() {
+	return CallNamedFunction("OnUpdateUI");
+}
+
 bool LuaExtension::OnMarginClick() {
 	return CallNamedFunction("OnMarginClick");
 }
