@@ -7,30 +7,12 @@
 
 extern const char appName[];
 
-extern const char pathSepString[];
-extern const char pathSepChar;
-extern const char configFileVisibilityString[];
 extern const char propUserFileName[];
 extern const char propGlobalFileName[];
 extern const char propAbbrevFileName[];
-extern const char fileRead[];
-extern const char fileWrite[];
 
 extern const char menuAccessIndicator[];
 
-#ifdef unix
-#include <limits.h>
-#ifdef PATH_MAX
-#define MAX_PATH PATH_MAX
-#else
-#define MAX_PATH 260
-#endif
-#endif
-
-#ifdef __vms
-const char *VMSToUnixStyle(const char *fileName);
-#endif
-
 #ifdef WIN32
 #ifdef _MSC_VER
 // Shut up level 4 warning:
@@ -108,35 +90,6 @@ public:
 	}
 };
 
-class PropSetFile : public PropSet {
-	bool lowerKeys;
-public:
-	PropSetFile(bool lowerKeys_=false);
-	~PropSetFile();
-	bool ReadLine(const char *data, bool ifIsTrue, const char *directoryForImports, SString imports[] = 0, int sizeImports = 0);
-	void ReadFromMemory(const char *data, int len, const char *directoryForImports, SString imports[] = 0, int sizeImports = 0);
-	bool Read(const char *filename, const char *directoryForImports, SString imports[] = 0, int sizeImports = 0);
-	void SetInteger(const char *key, sptr_t i);
-};
-
-class FilePath {
-	SString fileName;
-public:
-	FilePath() : fileName("") {
-	}
-	void Set(const char *fileName_) {
-		fileName = fileName_;
-	}
-	void Init() {
-		fileName = "";
-	}
-	bool SameNameAs(const char *other) const;
-	bool SameNameAs(const FilePath &other) const;
-	bool IsSet() const { return fileName.length() > 0; }
-	bool IsUntitled() const;
-	const char *FullPath() const;
-};
-
 class RecentFile : public FilePath {
 public:
 	CharacterRange selection;
@@ -276,11 +229,13 @@ public:
 	bool useMonoFont;
 	UniMode unicodeMode;
 	time_t fileModTime;
-	SString overrideExtension;
+	time_t fileModLastAsk;
+	enum { fmNone, fmMarked, fmModified} findMarks;
+	SString overrideExtension;	///< User has chosen to use a particular language
 	FoldState foldState;
 	Buffer() :
 		RecentFile(), doc(0), isDirty(false), useMonoFont(false),
-		unicodeMode(uni8Bit), fileModTime(0), foldState() {
+		unicodeMode(uni8Bit), fileModTime(0), fileModLastAsk(0), findMarks(fmNone), foldState() {
 	}
 
 	void Init() {
@@ -289,30 +244,47 @@ public:
 		useMonoFont = false;
 		unicodeMode = uni8Bit;
 		fileModTime = 0;
+		fileModLastAsk = 0;
+		findMarks = fmNone;
 		overrideExtension = "";
 		foldState.Clear();
 	}
+
+	void SetTimeFromFile() {
+		fileModTime = ModifiedTime();
+		fileModLastAsk = fileModTime;
+	}
 };
 
 class BufferList {
 protected:
 	int current;
+	int stackcurrent;
+	int *stack;
 public:
 	Buffer *buffers;
 	int size;
 	int length;
+	bool initialised;
 	BufferList();
 	~BufferList();
 	void Allocate(int maxSize);
 	int Add();
-	int GetDocumentByName(const char *filename);
+	int GetDocumentByName(FilePath filename);
 	void RemoveCurrent();
 	int Current();
+	Buffer *CurrentBuffer();
 	void SetCurrent(int index);
+	int StackNext();
+	int StackPrev();
+	void CommitStackSelection();
+	void MoveToStackTop(int index);
+private:
+	void PopStack();
 };
 
 enum JobSubsystem {
-    jobCLI = 0, jobGUI = 1, jobShell = 2, jobExtension = 3, jobHelp = 4, jobOtherHelp = 5};
+    jobCLI = 0, jobGUI = 1, jobShell = 2, jobExtension = 3, jobHelp = 4, jobOtherHelp = 5, jobGrep = 6};
 
 enum JobFlags {
     jobForceQueue=1, jobHasInput=2, jobQuiet=4, // 8 reserved for jobVeryQuiet
@@ -322,7 +294,7 @@ enum JobFlags {
 class Job {
 public:
 	SString command;
-	SString directory;
+	FilePath directory;
 	SString input;
 	JobSubsystem jobType;
 	int flags;
@@ -413,22 +385,16 @@ struct StyleAndWords {
 class SciTEBase : public ExtensionAPI {
 protected:
 	SString windowName;
-	char fullPath[MAX_PATH];
-	char fileName[MAX_PATH];
-	char fileExt[MAX_PATH];
-	char dirName[MAX_PATH];
-	SString dirNameAtExecute;
-	SString dirNameForExecute;
-	bool useMonoFont;
-	time_t fileModTime;
-	time_t fileModLastAsk;
+	FilePath filePath;
+	FilePath dirNameAtExecute;
+	FilePath dirNameForExecute;
 
 	enum { fileStackMax = 10 };
 	RecentFile recentFileStack[fileStackMax];
 	enum { fileStackCmdID = IDM_MRUFILE, bufferCmdID = IDM_BUFFER };
 
 	enum { importMax = 50 };
-	SString importFiles[importMax];
+	FilePath importFiles[importMax];
 	enum { importCmdID = IDM_IMPORT };
 
 	SString findWhat;
@@ -464,10 +430,8 @@ protected:
 
 	int codePage;
 	int characterSet;
-	UniMode unicodeMode;
 	SString language;
 	int lexLanguage;
-	SString overrideExtension;	///< User has chosen to use a particular language
 	WordList apis;
 	SString apisFileNames;
 	SString functionDefinition;
@@ -491,9 +455,6 @@ protected:
 	Window wEditor;
 	Window wOutput;
 	Window wIncrement;
-#if PLAT_GTK
-	Window wDivider;	// Not used on Windows
-#endif
 	Window wToolBar;
 	Window wStatusBar;
 	Window wTabBar;
@@ -579,11 +540,10 @@ protected:
 	bool usePalette;
 	bool clearBeforeExecute;
 	bool allowMenuActions;
-	bool isDirty;
 	bool isBuilding;
 	bool isBuilt;
 	bool executing;
-	bool scrollOutput;
+	int scrollOutput;
 	bool returnOutputToCommand;
 	enum { commandMax = 2 };
 	int commandCurrent;
@@ -604,6 +564,8 @@ protected:
 
 	PropSetFile propsAbbrev;
 
+	FilePath pathAbbreviations;
+
 	PropSetFile propsUI;
 
 	PropSetFile propsStatus;	// Not attached to a file but need SetInteger method.
@@ -617,13 +579,18 @@ protected:
 	void UpdateBuffersCurrent();
 	bool IsBufferAvailable();
 	bool CanMakeRoom(bool maySaveIfDirty=true);
-	void SetDocumentAt(int index);
+	void SetDocumentAt(int index, bool updateStack=true);
+	Buffer *CurrentBuffer() {
+		return buffers.CurrentBuffer();
+	}
 	void BuffersMenu();
 	void Next();
 	void Prev();
+	void NextInStack();
+	void PrevInStack();
+	void EndStackedTabbing();
 
 	void ReadGlobalPropFile();
-	void GetDocumentDirectory(char *docDir, int len);
 	void ReadAbbrevPropFile();
 	void ReadLocalPropFile();
 
@@ -650,9 +617,14 @@ protected:
 
 	virtual void WarnUser(int warnID) = 0;
 	void SetWindowName();
-	void SetFileName(const char *openName, bool fixCase = true);
+	void SetFileName(FilePath openName, bool fixCase = true);
+	FilePath FileNameExt() const {
+		return filePath.Name();
+	}
 	void ClearDocument();
+	void CreateBuffers();
 	void InitialiseBuffers();
+	FilePath RecentFilePath(const char *name);
 	void LoadRecentMenu();
 	void SaveRecentStack();
 	void LoadSession(const char *sessionName);
@@ -662,14 +634,13 @@ protected:
 	void New();
 	void Close(bool updateUI = true, bool loadingSession = false, bool makingRoomForNew = false);
 	bool IsAbsolutePath(const char *path);
-	bool Exists(const char *dir, const char *path, char *testPath);
+	bool Exists(const char *dir, const char *path, FilePath *resultPath);
 	void DiscoverEOLSetting();
 	void DiscoverIndentSetting();
+	SString DiscoverLanguage(const char *buf, size_t length);
 	void OpenFile(int fileSize, bool suppressMessage);
 	virtual void OpenUriList(const char *) {};
-	virtual void AbsolutePath(char *fullPath, const char *basePath, int size) = 0;
-	virtual void FixFilePath();
-	virtual bool OpenDialog(const char *filter=NULL) = 0;
+	virtual bool OpenDialog(FilePath directory, const char *filter) = 0;
 	virtual bool SaveAsDialog() = 0;
 	virtual void LoadSessionDialog() { };
 	virtual void SaveSessionDialog() { };
@@ -682,37 +653,35 @@ protected:
 		ofQuiet=8		// Avoid "Could not open file" message
 	};
 	virtual bool PreOpenCheck(const char *file);
-	bool Open(const char *file, OpenFlags of=ofNone);
+	bool Open(FilePath file, OpenFlags of=ofNone);
 	bool OpenSelected();
 	void Revert();
+	FilePath SaveName(const char *ext);
 	int SaveIfUnsure(bool forceQuestion = false);
 	int SaveIfUnsureAll(bool forceQuestion = false);
 	int SaveIfUnsureForBuilt();
 	bool Save();
-	virtual bool SaveAs(const char *file = 0);
+	void SaveAs(const char *file);
 	virtual void SaveACopy() = 0;
-	void SaveToHTML(const char *saveName);
+	void SaveToHTML(FilePath saveName);
 	void StripTrailingSpaces();
 	void EnsureFinalNewLine();
-	bool SaveBuffer(const char *saveName);
+	bool SaveBuffer(FilePath saveName);
 	virtual void SaveAsHTML() = 0;
-	void SaveToRTF(const char *saveName, int start = 0, int end = -1);
+	void SaveToRTF(FilePath saveName, int start = 0, int end = -1);
 	virtual void SaveAsRTF() = 0;
-	void SaveToPDF(const char *saveName);
+	void SaveToPDF(FilePath saveName);
 	virtual void SaveAsPDF() = 0;
-	void SaveToTEX(const char *saveName);
+	void SaveToTEX(FilePath saveName);
 	virtual void SaveAsTEX() = 0;
-	void SaveToXML(const char *saveName);
+	void SaveToXML(FilePath saveName);
 	virtual void SaveAsXML() = 0;
-	virtual void GetDefaultDirectory(char *directory, size_t size) = 0;
-	virtual bool GetSciteDefaultHome(char *path, unsigned int lenPath) = 0;
-	virtual bool GetSciteUserHome(char *path, unsigned int lenPath) = 0;
-	bool GetDefaultPropertiesFileName(char *pathDefaultProps,
-	        char *pathDefaultDir, unsigned int lenPath);
-	bool GetUserPropertiesFileName(char *pathUserProps,
-		char *pathUserDir, unsigned int lenPath);
-	bool GetAbbrevPropertiesFileName(char *pathAbbrevProps,
-	        char *pathDefaultDir, unsigned int lenPath);
+	virtual FilePath GetDefaultDirectory() = 0;
+	virtual FilePath GetSciteDefaultHome() = 0;
+	virtual FilePath GetSciteUserHome() = 0;
+	FilePath GetDefaultPropertiesFileName();
+	FilePath GetUserPropertiesFileName();
+	FilePath GetAbbrevPropertiesFileName();
 	void OpenProperties(int propsFile);
 	int GetMenuCommandAsInt(SString commandName);
 	virtual void Print(bool) {};
@@ -722,9 +691,10 @@ protected:
 	//	void SelectionExtend(char *sel, int len, char *notselchar);
 	void GetCTag(char *sel, int len);
 	virtual SString GetRangeInUIEncoding(Window &win, int selStart, int selEnd);
+	SString GetLine(Window &win, int line);
 	SString RangeExtendAndGrab(Window &wCurrent, int &selStart, int &selEnd,
-		bool (*ischarforsel)(char ch), bool stripEol = true);
-	SString SelectionExtend(bool (*ischarforsel)(char ch), bool stripEol = true);
+		bool (SciTEBase::*ischarforsel)(char ch), bool stripEol = true);
+	SString SelectionExtend(bool (SciTEBase::*ischarforsel)(char ch), bool stripEol = true);
 	void FindWordAtCaret(int &start, int &end);
 	bool SelectWordAtCaret();
 	SString SelectionWord(bool stripEol = true);
@@ -734,7 +704,7 @@ protected:
 	virtual SString EncodeString(const SString &s);
 	virtual void Find() = 0;
 	virtual int WindowMessageBox(Window &w, const SString &m, int style)=0;
-	void FindMessageBox(const SString &msg);
+	virtual void FindMessageBox(const SString &msg, const SString *findItem=0)=0;
 	int FindInTarget(const char *findWhat, int lenFind, int startPosition, int endPosition);
 	int FindNext(bool reverseDirection, bool showWarnings = true);
 	virtual void FindIncrement() = 0;
@@ -756,8 +726,8 @@ protected:
 	void GoMatchingBrace(bool select);
 	void GoMatchingPreprocCond(int direction, bool select);
 	virtual void FindReplace(bool replace) = 0;
-	void OutputAppendString(const char *s, int len = -1, bool startLine = false);
-	void OutputAppendStringSynchronised(const char *s, int len = -1, bool startLine = false);
+	void OutputAppendString(const char *s, int len = -1);
+	void OutputAppendStringSynchronised(const char *s, int len = -1);
 	void MakeOutputVisible();
 	void ClearJobQueue();
 	virtual void Execute();
@@ -784,6 +754,7 @@ protected:
 	void MaintainIndentation(char ch);
 	void AutomaticIndentation(char ch);
 	void CharAdded(char ch);
+	void CharAddedOutput(int ch);
 	void SetTextProperties(PropSetFile &ps);
 	virtual void SetFileProperties(PropSet &ps) = 0;
 	virtual void UpdateStatusBar(bool bUpdateSlowData);
@@ -795,6 +766,7 @@ protected:
 		int flags = 0);
 	virtual void AboutDialog() = 0;
 	virtual void QuitProgram() = 0;
+	void CloseTab(int tab);
 	void CloseAllBuffers(bool loadingSession = false);
 	int SaveAllBuffers(bool forceQuestion, bool alwaysYes=false);
 	void SaveTitledBuffers();
@@ -818,6 +790,7 @@ protected:
 	virtual void ShowTabBar() = 0;
 	virtual void ShowStatusBar() = 0;
 
+	void RemoveFindMarks();
 	int MarkAll();
 	void BookmarkAdd(int lineno = -1);
 	void BookmarkDelete(int lineno = -1);
@@ -841,9 +814,9 @@ protected:
 	void DeleteFileStackMenu();
 	void SetFileStackMenu();
 	void DropFileStackTop();
-	void AddFileToBuffer(const char *file, int pos);
-	void AddFileToStack(const char *file, CharacterRange selection, int scrollPos);
-	void RemoveFileFromStack(const char *file);
+	void AddFileToBuffer(FilePath file, int pos);
+	void AddFileToStack(FilePath file, CharacterRange selection, int scrollPos);
+	void RemoveFileFromStack(FilePath file);
 	RecentFile GetFilePosition();
 	void DisplayAround(const RecentFile &rf);
 	void StackMenu(int pos);
@@ -903,6 +876,12 @@ protected:
 	void ContinueMacroList(const char *stxt);
 	void LoadMRUAndSession(bool allowLoadSession);
 	bool ProcessCommandLine(SString &args, int phase);
+	virtual bool IsStdinBlocked();
+	void OpenFromStdin(bool UseOutputPane);
+	void OpenFilesFromStdin();
+	enum GrepFlags { grepNone=0, grepWholeWord=1, grepMatchCase=2, grepStdOut=4 };
+	void GrepRecursive(GrepFlags gf, FilePath baseDir, const char *searchString, const char *fileTypes);
+	void InternalGrep(GrepFlags gf, const char *directory, const char *files, const char *search);
 	void EnumProperties(const char *action);
 	void SendOneProperty(const char *kind, const char *key, const char *val);
 	void PropertyFromDirector(const char *arg);
@@ -921,6 +900,10 @@ protected:
 	void ShutDown();
 	void Perform(const char *actions);
 
+	// Valid CurrentWord characters
+	bool iswordcharforsel(char ch);
+	bool isfilenamecharforsel(char ch);
+	bool islexerwordcharforsel(char ch);
 public:
 
 	SciTEBase(Extension *ext = 0);
@@ -952,13 +935,8 @@ const int blockSize = 131072;
 #endif
 
 int ControlIDOfCommand(unsigned long);
-bool BuildPath(char *path, const char *dir, const char *fileName,
-	unsigned int lenPath);
-time_t GetModTime(const char *fullPath);
-bool IsUntitledFileName(const char *name);
 void LowerCaseString(char *s);
-void ChopTerminalSlash(char *path);
 long ColourOfProperty(PropSet &props, const char *key, ColourDesired colourDefault);
-char *Slash(const char *s);
+char *Slash(const char *s, bool quoteQuotes);
 unsigned int UnSlash(char *s);
 void WindowSetFocus(Window &w);
