@@ -55,6 +55,8 @@
 #include "Scintilla.h"
 #include "SciLexer.h"
 #include "Extender.h"
+#include "FilePath.h"
+#include "PropSetFile.h"
 #include "SciTEBase.h"
 
 const char recentFileName[] = "SciTE.recent";
@@ -66,37 +68,17 @@ Job::Job() {
 
 void Job::Clear() {
 	command = "";
-	directory = "";
+	directory.Init();
 	jobType = jobCLI;
 	input = "";
 	flags = 0;
 }
 
-bool FilePath::SameNameAs(const char *other) const {
-#ifdef WIN32
-	return EqualCaseInsensitive(fileName.c_str(), other);
-#else
-
-	return fileName == other;
-#endif
-}
-
-bool FilePath::SameNameAs(const FilePath &other) const {
-	return SameNameAs(other.fileName.c_str());
-}
-
-bool FilePath::IsUntitled() const {
-	return IsUntitledFileName(fileName.c_str());
-}
-
-const char *FilePath::FullPath() const {
-	return fileName.c_str();
-}
-
-BufferList::BufferList() : current(0), buffers(0), size(0), length(0) {}
+BufferList::BufferList() : current(0), stackcurrent(0), buffers(0), size(0), length(0), initialised(false) {}
 
 BufferList::~BufferList() {
 	delete []buffers;
+	delete []stack;
 }
 
 void BufferList::Allocate(int maxSize) {
@@ -104,6 +86,8 @@ void BufferList::Allocate(int maxSize) {
 	current = 0;
 	size = maxSize;
 	buffers = new Buffer[size];
+	stack = new int[size];
+	stack[0] = 0;
 }
 
 int BufferList::Add() {
@@ -111,12 +95,14 @@ int BufferList::Add() {
 		length++;
 	}
 	buffers[length - 1].Init();
+	stack[length - 1] = length - 1;
+	MoveToStackTop(length - 1);
 
 	return length - 1;
 }
 
-int BufferList::GetDocumentByName(const char *filename) {
-	if (!filename || !filename[0]) {
+int BufferList::GetDocumentByName(FilePath filename) {
+	if (!filename.IsSet()) {
 		return -1;
 	}
 	for (int i = 0;i < length;i++) {
@@ -136,6 +122,8 @@ void BufferList::RemoveCurrent() {
 	buffers[length - 1].doc = currentDoc;
 
 	if (length > 1) {
+		CommitStackSelection();
+		PopStack();
 		length--;
 
 		buffers[length].Init();
@@ -148,16 +136,62 @@ void BufferList::RemoveCurrent() {
 	} else {
 		buffers[current].Init();
 	}
+	MoveToStackTop(current);
 }
 
 int BufferList::Current() {
 	return current;
 }
 
+Buffer *BufferList::CurrentBuffer() {
+	return &buffers[Current()];
+}
+
 void BufferList::SetCurrent(int index) {
 	current = index;
 }
 
+void BufferList::PopStack() {
+	for (int i=0; i<length-1; ++i) {
+		int index = stack[i + 1];
+		// adjust the index for items that will move in buffers[]
+		if (index > current)
+			--index;
+		stack[i] = index;
+	}
+}
+
+int BufferList::StackNext() {
+	if (++stackcurrent >= length)
+		stackcurrent = 0;
+	return stack[stackcurrent];
+}
+
+int BufferList::StackPrev() {
+	if (--stackcurrent < 0)
+		stackcurrent = length - 1;
+	return stack[stackcurrent];
+}
+
+void BufferList::MoveToStackTop(int index) {
+	// shift top chunk of stack down into the slot that index occupies
+	bool move = false;
+	for (int i=length-1; i>0; --i) {
+		if (stack[i] == index)
+			move = true;
+		if (move)
+			stack[i] = stack[i-1];
+	}
+	stack[0] = index;
+}
+
+void BufferList::CommitStackSelection() {
+	// called only when ctrl key is released when ctrl-tabbing
+	// or when a document is closed (in case of Ctrl+F4 during ctrl-tabbing)
+	MoveToStackTop(stack[stackcurrent]);
+	stackcurrent = 0;
+}
+
 sptr_t SciTEBase::GetDocumentAt(int index) {
 	if (index < 0 || index >= buffers.size) {
 		//Platform::DebugPrintf("SciTEBase::GetDocumentAt: Index out of range.\n");
@@ -170,7 +204,7 @@ sptr_t SciTEBase::GetDocumentAt(int index) {
 	return buffers.buffers[index].doc;
 }
 
-void SciTEBase::SetDocumentAt(int index) {
+void SciTEBase::SetDocumentAt(int index, bool updateStack) {
 	int currentbuf = buffers.Current();
 
 	if (	index < 0 ||
@@ -183,6 +217,9 @@ void SciTEBase::SetDocumentAt(int index) {
 	UpdateBuffersCurrent();
 
 	buffers.SetCurrent(index);
+	if (updateStack) {
+		buffers.MoveToStackTop(index);
+	}
 
 	if (extender) {
 		if (buffers.size > 1)
@@ -192,16 +229,11 @@ void SciTEBase::SetDocumentAt(int index) {
 	}
 
 	Buffer bufferNext = buffers.buffers[buffers.Current()];
-	isDirty = bufferNext.isDirty;
-	useMonoFont = bufferNext.useMonoFont;
-	unicodeMode = bufferNext.unicodeMode;
-	fileModTime = bufferNext.fileModTime;
-	overrideExtension = bufferNext.overrideExtension;
-	SetFileName(bufferNext.FullPath());
+	SetFileName(bufferNext);
 	SendEditor(SCI_SETDOCPOINTER, 0, GetDocumentAt(buffers.Current()));
 	SetWindowName();
 	ReadProperties();
-	if (unicodeMode != uni8Bit) {
+	if (CurrentBuffer()->unicodeMode != uni8Bit) {
 		// Override the code page if Unicode
 		codePage = SC_CP_UTF8;
 		SendEditor(SCI_SETCODEPAGE, codePage);
@@ -244,7 +276,7 @@ void SciTEBase::SetDocumentAt(int index) {
 	UpdateStatusBar(true);
 
 	if (extender) {
-		extender->OnSwitchFile(fullPath);
+		extender->OnSwitchFile(filePath.AsFileSystem());
 	}
 }
 
@@ -252,14 +284,9 @@ void SciTEBase::UpdateBuffersCurrent() {
 	int currentbuf = buffers.Current();
 
 	if ((buffers.length > 0) && (currentbuf >= 0)) {
-		buffers.buffers[currentbuf].Set(fullPath);
+		buffers.buffers[currentbuf].Set(filePath);
 		buffers.buffers[currentbuf].selection = GetSelection();
 		buffers.buffers[currentbuf].scrollPosition = GetCurrentScrollPosition();
-		buffers.buffers[currentbuf].isDirty = isDirty;
-		buffers.buffers[currentbuf].useMonoFont = useMonoFont;
-		buffers.buffers[currentbuf].fileModTime = fileModTime;
-		buffers.buffers[currentbuf].overrideExtension = overrideExtension;
-		buffers.buffers[currentbuf].unicodeMode = unicodeMode;
 
 		// retrieve fold state and store in buffer state info
 		int maxLine = SendEditor(SCI_GETLINECOUNT);
@@ -312,32 +339,31 @@ bool SciTEBase::CanMakeRoom(bool maySaveIfDirty) {
 	return false;
 }
 
-bool IsUntitledFileName(const char *name) {
-	const char *dirEnd = strrchr(name, pathSepChar);
-	return !dirEnd || !dirEnd[1];
-}
-
 void SciTEBase::ClearDocument() {
 	SendEditor(SCI_CLEARALL);
 	SendEditor(SCI_EMPTYUNDOBUFFER);
 	SendEditor(SCI_SETSAVEPOINT);
 }
 
+void SciTEBase::CreateBuffers() {
+	int buffersWanted = props.GetInt("buffers");
+	if (buffersWanted > bufferMax) {
+		buffersWanted = bufferMax;
+	}
+	if (buffersWanted < 1) {
+		buffersWanted = 1;
+	}
+	buffers.Allocate(buffersWanted);
+}
+
 void SciTEBase::InitialiseBuffers() {
-	if (buffers.size == 0) {
-		int buffersWanted = props.GetInt("buffers");
-		if (buffersWanted > bufferMax) {
-			buffersWanted = bufferMax;
-		}
-		if (buffersWanted < 1) {
-			buffersWanted = 1;
-		}
-		buffers.Allocate(buffersWanted);
+	if (!buffers.initialised) {
+		buffers.initialised = true;
 		// First document is the default from creation of control
 		buffers.buffers[0].doc = SendEditor(SCI_GETDOCPOINTER, 0, 0);
 		SendEditor(SCI_ADDREFDOCUMENT, 0, buffers.buffers[0].doc); // We own this reference
-		if (buffersWanted == 1) {
-			// No buffers, delete the Buffers main menu entry
+		if (buffers.size == 1) {
+			// Single buffer mode, delete the Buffers main menu entry
 			DestroyMenuItem(menuBuffers, 0);
 #if PLAT_WIN
 			// Make previous change visible.
@@ -350,40 +376,15 @@ void SciTEBase::InitialiseBuffers() {
 	}
 }
 
-static void EnsureEndsWithPathSeparator(char *path) {
-	size_t pathLen = strlen(path);
-	if ((pathLen > 0) && path[pathLen - 1] != pathSepChar) {
-		strcat(path, pathSepString);
-	}
-}
-
-static void RecentFilePath(char *path, const char *name) {
-	char *where = getenv("SciTE_HOME");
-	if (!where)
-		where = getenv("HOME");
-	if (!where) {
-#if PLAT_WIN
-		::GetModuleFileName(0, path, MAX_PATH);
-		char *lastSlash = strrchr(path, pathSepChar);
-		if (lastSlash)
-			*(lastSlash + 1) = '\0';
-#else
-
-		*path = '\0';
-#endif
-
-	} else {
-		strcpy(path, where);
-		EnsureEndsWithPathSeparator(path);
-	}
-	strcat(path, configFileVisibilityString);
-	strcat(path, name);
+FilePath SciTEBase::RecentFilePath(const char *name) {
+	SString nameWithVisibility(configFileVisibilityString);
+	nameWithVisibility.append(name);
+	return FilePath(GetSciteUserHome(), nameWithVisibility.c_str());
 }
 
 void SciTEBase::LoadRecentMenu() {
-	char recentPathName[MAX_PATH + 1];
-	RecentFilePath(recentPathName, recentFileName);
-	FILE *recentFile = fopen(recentPathName, fileRead);
+	FilePath recentPathName = RecentFilePath(recentFileName);
+	FILE *recentFile = recentPathName.Open(fileRead);
 	if (!recentFile) {
 		DeleteFileStackMenu();
 		return;
@@ -401,33 +402,32 @@ void SciTEBase::LoadRecentMenu() {
 }
 
 void SciTEBase::SaveRecentStack() {
-	char recentPathName[MAX_PATH + 1];
-	RecentFilePath(recentPathName, recentFileName);
-	FILE *recentFile = fopen(recentPathName, fileWrite);
+	FilePath recentPathName = RecentFilePath(recentFileName);
+	FILE *recentFile = recentPathName.Open(fileWrite);
 	if (!recentFile)
 		return;
 	int i;
 	// save recent files list
 	for (i = fileStackMax - 1; i >= 0; i--) {
 		if (recentFileStack[i].IsSet())
-			fprintf(recentFile, "%s\n", recentFileStack[i].FullPath());
+			fprintf(recentFile, "%s\n", recentFileStack[i].AsInternal());
 	}
 	// save buffers list
 	for (i = buffers.length - 1; i >= 0; i--) {
 		if (buffers.buffers[i].IsSet())
-			fprintf(recentFile, "%s\n", buffers.buffers[i].FullPath());
+			fprintf(recentFile, "%s\n", buffers.buffers[i].AsInternal());
 	}
 	fclose(recentFile);
 }
 
 void SciTEBase::LoadSession(const char *sessionName) {
-	char sessionPathName[MAX_PATH + 1];
+	FilePath sessionPathName;
 	if (sessionName[0] == '\0') {
-		RecentFilePath(sessionPathName, defaultSessionFileName);
+		sessionPathName = RecentFilePath(defaultSessionFileName);
 	} else {
-		strcpy(sessionPathName, sessionName);
+		sessionPathName.Set(sessionName);
 	}
-	FILE *sessionFile = fopen(sessionPathName, fileRead);
+	FILE *sessionFile = sessionPathName.Open(fileRead);
 	if (!sessionFile)
 		return;
 	// comment next line if you don't want to close all buffers before loading session
@@ -435,9 +435,9 @@ void SciTEBase::LoadSession(const char *sessionName) {
 	int curr = -1, pos = 0;
 	char *file, line[MAX_PATH + 128];
 	for (int i = 0; i < bufferMax; i++) {
-		if (!fgets (line, sizeof (line), sessionFile))
+		if (!fgets(line, sizeof (line), sessionFile))
 			break;
-		line[strlen (line) - 1] = '\0';
+		line[strcspn(line, "\r\n")] = '\0';	// Remove \r, \n, or \r\n.
 		if (sscanf(line, "<pos=%i>", &pos) != 1)
 			break;
 		file = strchr(line, '>') + 2;
@@ -450,23 +450,22 @@ void SciTEBase::LoadSession(const char *sessionName) {
 	}
 	fclose(sessionFile);
 
-	if (!IsAbsolutePath(sessionPathName))
-		AbsolutePath(sessionPathName, sessionPathName, MAX_PATH);
+	FilePath sessionFilePath = FilePath(sessionPathName).AbsolutePath();
 	// add/update SessionPath environment variable
-	props.Set("SessionPath", sessionPathName);
+	props.Set("SessionPath", sessionFilePath.AsFileSystem());
 
 	if (curr != -1)
 		SetDocumentAt(curr);
 }
 
 void SciTEBase::SaveSession(const char *sessionName) {
-	char sessionPathName[MAX_PATH + 1];
+	FilePath sessionPathName;
 	if (sessionName[0] == '\0') {
-		RecentFilePath(sessionPathName, defaultSessionFileName);
+		sessionPathName = RecentFilePath(defaultSessionFileName);
 	} else {
-		strcpy(sessionPathName, sessionName);
+		sessionPathName.Set(sessionName);
 	}
-	FILE *sessionFile = fopen(sessionPathName, fileWrite);
+	FILE *sessionFile = sessionPathName.Open(fileWrite);
 	if (!sessionFile)
 		return;
 	int curr = buffers.Current();
@@ -477,15 +476,14 @@ void SciTEBase::SaveSession(const char *sessionName) {
 			pos = SendEditor(SCI_GETCURRENTPOS) + 1;
 			if (i == curr)
 				pos = -pos;
-			fprintf(sessionFile, "<pos=%i> %s\n", pos, buffers.buffers[i].FullPath());
+			fprintf(sessionFile, "<pos=%i> %s\n", pos, buffers.buffers[i].AsInternal());
 		}
 	}
 	fclose(sessionFile);
 
-	if (!IsAbsolutePath(sessionPathName))
-		AbsolutePath(sessionPathName, sessionPathName, MAX_PATH);
+	FilePath sessionFilePath = FilePath(sessionPathName).AbsolutePath();
 	// add/update SessionPath environment variable
-	props.Set("SessionPath", sessionPathName);
+	props.Set("SessionPath", sessionFilePath.AsFileSystem());
 
 	SetDocumentAt(curr);
 }
@@ -536,7 +534,7 @@ void SciTEBase::New() {
 	UpdateBuffersCurrent();
 
 	if ((buffers.size == 1) && (!buffers.buffers[0].IsUntitled())) {
-		AddFileToStack(buffers.buffers[0].FullPath(),
+		AddFileToStack(buffers.buffers[0],
 		               buffers.buffers[0].selection,
 		               buffers.buffers[0].scrollPosition);
 	}
@@ -556,12 +554,10 @@ void SciTEBase::New() {
 	sptr_t doc = GetDocumentAt(buffers.Current());
 	SendEditor(SCI_SETDOCPOINTER, 0, doc);
 
-	fullPath[0] = '\0';
-	fileName[0] = '\0';
-	fileExt[0] = '\0';
-	dirName[0] = '\0';
-	SetFileName(fileName);
-	isDirty = false;
+	FilePath curDirectory(filePath.Directory());
+	filePath.Set(curDirectory, "");
+	SetFileName(filePath);
+	CurrentBuffer()->isDirty = false;
 	isBuilding = false;
 	isBuilt = false;
 	isReadOnly = false;	// No sense to create an empty, read-only buffer...
@@ -580,8 +576,7 @@ void SciTEBase::Close(bool updateUI, bool loadingSession, bool makingRoomForNew)
 		// With no buffer list, Close means close from MRU
 		closingLast = !(recentFileStack[0].IsSet());
 		buffers.buffers[0].Init();
-		buffers.buffers[0].useMonoFont = useMonoFont;
-		fullPath[0] = '\0';
+		filePath.Set("");
 		ClearDocument(); //avoid double are-you-sure
 		if (!makingRoomForNew)
 			StackMenu(0); // calls New, or Open, which calls InitBuffer
@@ -589,37 +584,33 @@ void SciTEBase::Close(bool updateUI, bool loadingSession, bool makingRoomForNew)
 		if (buffers.Current() >= 0 && buffers.Current() < buffers.length) {
 			UpdateBuffersCurrent();
 			Buffer buff = buffers.buffers[buffers.Current()];
-			AddFileToStack(buff.FullPath(), buff.selection, buff.scrollPosition);
+			AddFileToStack(buff, buff.selection, buff.scrollPosition);
 		}
 		closingLast = (buffers.length == 1);
 		if (closingLast) {
 			buffers.buffers[0].Init();
-			buffers.buffers[0].useMonoFont = useMonoFont;
 			if (extender)
 				extender->InitBuffer(0);
 		} else {
-			extender->RemoveBuffer(buffers.Current());
+			if (extender)
+				extender->RemoveBuffer(buffers.Current());
 			buffers.RemoveCurrent();
 			if (extender && !makingRoomForNew)
 				extender->ActivateBuffer(buffers.Current());
 		}
 		Buffer bufferNext = buffers.buffers[buffers.Current()];
-		isDirty = bufferNext.isDirty;
-		useMonoFont = bufferNext.useMonoFont;
-		unicodeMode = bufferNext.unicodeMode;
-		fileModTime = bufferNext.fileModTime;
-		overrideExtension = bufferNext.overrideExtension;
 
 		if (updateUI)
-			SetFileName(bufferNext.FullPath());
+			SetFileName(bufferNext);
 		SendEditor(SCI_SETDOCPOINTER, 0, GetDocumentAt(buffers.Current()));
 		if (closingLast) {
+			SendEditor(SCI_SETREADONLY, 0);
 			ClearDocument();
 		}
 		if (updateUI) {
 			SetWindowName();
 			ReadProperties();
-			if (unicodeMode != uni8Bit) {
+			if (CurrentBuffer()->unicodeMode != uni8Bit) {
 				// Override the code page if Unicode
 				codePage = SC_CP_UTF8;
 				SendEditor(SCI_SETCODEPAGE, codePage);
@@ -651,11 +642,33 @@ void SciTEBase::Close(bool updateUI, bool loadingSession, bool makingRoomForNew)
 		UpdateStatusBar(true);
 	}
 
+	if (extender && !closingLast) {
+		extender->OnSwitchFile(filePath.AsFileSystem());
+	}
+
 	if (closingLast && props.GetInt("quit.on.close.last") && !loadingSession) {
 		QuitProgram();
 	}
 }
 
+void SciTEBase::CloseTab(int tab) {
+	int tabCurrent = buffers.Current();
+	if (tab == tabCurrent) {
+		if (SaveIfUnsure() != IDCANCEL) {
+			Close();
+		}
+	} else {
+		FilePath fpCurrent = buffers.buffers[tabCurrent].AbsolutePath();
+		SetDocumentAt(tab);
+		if (SaveIfUnsure() != IDCANCEL) {
+			Close();
+			WindowSetFocus(wEditor);
+			// Return to the previous buffer
+			SetDocumentAt(buffers.GetDocumentByName(fpCurrent));
+		}
+	}
+}
+
 void SciTEBase::CloseAllBuffers(bool loadingSession) {
 	if (SaveAllBuffers(false) != IDCANCEL) {
 		while (buffers.length > 1)
@@ -667,7 +680,7 @@ void SciTEBase::CloseAllBuffers(bool loadingSession) {
 
 int SciTEBase::SaveAllBuffers(bool forceQuestion, bool alwaysYes) {
 	int choice = IDYES;
-	UpdateBuffersCurrent();	// Ensure isDirty copied
+	UpdateBuffersCurrent();
 	int currentBuffer = buffers.Current();
 	for (int i = 0; (i < buffers.length) && (choice != IDCANCEL); i++) {
 		if (buffers.buffers[i].isDirty) {
@@ -686,7 +699,7 @@ int SciTEBase::SaveAllBuffers(bool forceQuestion, bool alwaysYes) {
 }
 
 void SciTEBase::SaveTitledBuffers() {
-	UpdateBuffersCurrent(); // Ensure isDirty copied
+	UpdateBuffersCurrent();
 	int currentBuffer = buffers.Current();
 	for (int i = 0; i < buffers.length; i++) {
 		if (buffers.buffers[i].isDirty && !buffers.buffers[i].IsUntitled()) {
@@ -714,6 +727,20 @@ void SciTEBase::Prev() {
 	CheckReload();
 }
 
+void SciTEBase::NextInStack() {
+	SetDocumentAt(buffers.StackNext(), false);
+	CheckReload();
+}
+
+void SciTEBase::PrevInStack() {
+	SetDocumentAt(buffers.StackPrev(), false);
+	CheckReload();
+}
+
+void SciTEBase::EndStackedTabbing() {
+	buffers.CommitStackSelection();
+}
+
 void SciTEBase::BuffersMenu() {
 	UpdateBuffersCurrent();
 	DestroyMenuItem(menuBuffers, IDM_BUFFERSEP);
@@ -757,7 +784,7 @@ void SciTEBase::BuffersMenu() {
 				strcat(entry, untitled.c_str());
 				strcat(titleTab, untitled.c_str());
 			} else {
-				SString path = buffers.buffers[pos].FullPath();
+				SString path = buffers.buffers[pos].AsInternal();
 #if PLAT_WIN
 				// Handle '&' characters in path, since they are interpreted in
 				// menues and tab names.
@@ -806,7 +833,7 @@ void SciTEBase::BuffersMenu() {
 				if (buffers.buffers[pos].IsUntitled())
 					tabcontent = gtk_label_new(LocaliseString("Untitled").c_str());
 				else
-					tabcontent = gtk_label_new(buffers.buffers[pos].FullPath());
+					tabcontent = gtk_label_new(buffers.buffers[pos].AsInternal());
 
 				gtk_widget_show(tablabel);
 				gtk_widget_show(tabcontent);
@@ -850,7 +877,7 @@ void SciTEBase::SetFileStackMenu() {
 				sprintf(entry, "&%d ", (stackPos + 1) % 10);
 #endif
 
-				strcat(entry, recentFileStack[stackPos].FullPath());
+				strcat(entry, recentFileStack[stackPos].AsInternal());
 				SetMenuItem(menuFile, MRU_START + stackPos + 1, itemID, entry);
 			}
 		}
@@ -865,21 +892,20 @@ void SciTEBase::DropFileStackTop() {
 	SetFileStackMenu();
 }
 
-void SciTEBase::AddFileToBuffer(const char *file, int pos) {
-	FILE *fp = fopen(file, fileRead);  // file existence test
-	if (fp) {                      // for missing files Open() gives an empty buffer - do not want this
-		fclose(fp);
+void SciTEBase::AddFileToBuffer(FilePath file, int pos) {
+	// file existence test
+	if (file.Exists()) {                      // for missing files Open() gives an empty buffer - do not want this
 		Open(file, ofForceLoad);
 		SendEditor(SCI_GOTOPOS, pos);
 	}
 }
 
-void SciTEBase::AddFileToStack(const char *file, CharacterRange selection, int scrollPos) {
-	if (!file)
+void SciTEBase::AddFileToStack(FilePath file, CharacterRange selection, int scrollPos) {
+	if (!file.IsSet())
 		return;
 	DeleteFileStackMenu();
 	// Only stack non-empty names
-	if ((file[0]) && (file[strlen(file) - 1] != pathSepChar)) {
+	if (file.IsSet() && !file.IsUntitled()) {
 		int stackPos;
 		int eqPos = fileStackMax - 1;
 		for (stackPos = 0; stackPos < fileStackMax; stackPos++)
@@ -894,8 +920,8 @@ void SciTEBase::AddFileToStack(const char *file, CharacterRange selection, int s
 	SetFileStackMenu();
 }
 
-void SciTEBase::RemoveFileFromStack(const char *file) {
-	if (!file || !file[0])
+void SciTEBase::RemoveFileFromStack(FilePath file) {
+	if (!file.IsSet())
 		return;
 	DeleteFileStackMenu();
 	int stackPos;
@@ -1001,9 +1027,8 @@ void SciTEBase::StackMenu(int pos) {
 			} else if (recentFileStack[pos].IsSet()) {
 				RecentFile rf = recentFileStack[pos];
 				//Platform::DebugPrintf("Opening pos %d %s\n",recentFileStack[pos].lineNumber,recentFileStack[pos].fileName);
-				overrideExtension = "";
 				// Already asked user so don't allow Open to ask again.
-				Open(rf.FullPath(), ofNoSaveIfDirty);
+				Open(rf, ofNoSaveIfDirty);
 				DisplayAround(rf);
 			}
 		}
@@ -1032,13 +1057,13 @@ void SciTEBase::SetToolsMenu() {
 		SString prefix = "command.name.";
 		prefix += SString(item);
 		prefix += ".";
-		SString commandName = props.GetNewExpand(prefix.c_str(), fileName);
+		SString commandName = props.GetNewExpand(prefix.c_str(), FileNameExt().AsInternal());
 		if (commandName.length()) {
 			SString sMenuItem = commandName;
 			prefix = "command.shortcut.";
 			prefix += SString(item);
 			prefix += ".";
-			SString sMnemonic = props.GetNewExpand(prefix.c_str(), fileName);
+			SString sMnemonic = props.GetNewExpand(prefix.c_str(), FileNameExt().AsInternal());
 			if (item < 10 && sMnemonic.length() == 0) {
 				sMnemonic += "Ctrl+";
 				sMnemonic += SString(item);
@@ -1087,7 +1112,7 @@ JobSubsystem SciTEBase::SubsystemType(const char *cmd, int item) {
 		subsysprefix += SString(item);
 		subsysprefix += ".";
 	}
-	SString subsystem = props.GetNewExpand(subsysprefix.c_str(), fileName);
+	SString subsystem = props.GetNewExpand(subsysprefix.c_str(), FileNameExt().AsInternal());
 	return SubsystemType(subsystem[0]);
 }
 
@@ -1100,7 +1125,7 @@ void SciTEBase::ToolsMenu(int item) {
 	SString propName = "command.";
 	propName += itemSuffix;
 
-	SString command = props.GetWild(propName.c_str(), fileName);
+	SString command = props.GetWild(propName.c_str(), FileNameExt().AsInternal());
 	if (command.length()) {
 		int saveBefore = 0;
 
@@ -1112,7 +1137,7 @@ void SciTEBase::ToolsMenu(int item) {
 
 		propName = "command.mode.";
 		propName += itemSuffix;
-		SString modeVal = props.GetNewExpand(propName.c_str(), fileName);
+		SString modeVal = props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal());
 		modeVal.remove(" ");
 		if (modeVal.length()) {
 			char *modeTags = modeVal.detach();
@@ -1200,45 +1225,45 @@ void SciTEBase::ToolsMenu(int item) {
 
 		propName = "command.save.before.";
 		propName += itemSuffix;
-		if (props.GetWild(propName.c_str(), fileName).length())
-			saveBefore = props.GetNewExpand(propName.c_str(), fileName).value();
+		if (props.GetWild(propName.c_str(), FileNameExt().AsInternal()).length())
+			saveBefore = props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal()).value();
 
-		if (saveBefore == 2 || (saveBefore == 1 && Save()) || SaveIfUnsure() != IDCANCEL) {
+		if (saveBefore == 2 || (saveBefore == 1 && (!(CurrentBuffer()->isDirty) || Save())) || SaveIfUnsure() != IDCANCEL) {
 			int flags = 0;
 
 			propName = "command.is.filter.";
 			propName += itemSuffix;
-			if (props.GetWild(propName.c_str(), fileName).length())
-				filter = (props.GetNewExpand(propName.c_str(), fileName)[0]=='1');
+			if (props.GetWild(propName.c_str(), FileNameExt().AsInternal()).length())
+				filter = (props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal())[0]=='1');
 			if (filter)
-				fileModTime -= 1;
+				CurrentBuffer()->fileModTime -= 1;
 
 			propName = "command.subsystem.";
 			propName += itemSuffix;
-			if (props.GetWild(propName.c_str(), fileName).length()) {
-				SString subsystemVal = props.GetNewExpand(propName.c_str(), fileName);
+			if (props.GetWild(propName.c_str(), FileNameExt().AsInternal()).length()) {
+				SString subsystemVal = props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal());
 				jobType = SubsystemType(subsystemVal[0]);
 			}
 
 			propName = "command.input.";
 			propName += itemSuffix;
 			SString input;
-			if (props.GetWild(propName.c_str(), fileName).length()) {
-				input = props.GetNewExpand(propName.c_str(), fileName);
+			if (props.GetWild(propName.c_str(), FileNameExt().AsInternal()).length()) {
+				input = props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal());
 				flags |= jobHasInput;
 			}
 
 			propName = "command.quiet.";
 			propName += itemSuffix;
-			if (props.GetWild(propName.c_str(), fileName).length())
-				quiet = (props.GetNewExpand(propName.c_str(), fileName).value()==1);
+			if (props.GetWild(propName.c_str(), FileNameExt().AsInternal()).length())
+				quiet = (props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal()).value()==1);
 			if (quiet)
 				flags |= jobQuiet;
 
 			propName = "command.replace.selection.";
 			propName += itemSuffix;
-			if (props.GetWild(propName.c_str(), fileName).length())
-				repSel = props.GetNewExpand(propName.c_str(), fileName).value();
+			if (props.GetWild(propName.c_str(), FileNameExt().AsInternal()).length())
+				repSel = props.GetNewExpand(propName.c_str(), FileNameExt().AsInternal()).value();
 
 			if (repSel == 1)
 				flags |= jobRepSelYes;
@@ -1299,10 +1324,14 @@ int DecodeMessage(char *cdoc, char *sourcePath, int format, int &column) {
 		}
 	case SCE_ERR_MS: {
 			// Visual *
-			char *endPath = strchr(cdoc, '(');
-			int length = endPath - cdoc;
+			char *start = cdoc;
+			while (isspacechar(*start)) {
+				start++;
+			}
+			char *endPath = strchr(start, '(');
+			int length = endPath - start;
 			if ((length > 0) && (length < MAX_PATH)) {
-				strncpy(sourcePath, cdoc, length);
+				strncpy(sourcePath, start, length);
 				sourcePath[length] = 0;
 			}
 			endPath++;
@@ -1382,10 +1411,11 @@ int DecodeMessage(char *cdoc, char *sourcePath, int format, int &column) {
 			char *line = strstr(cdoc, idLine);
 			char *file = strstr(cdoc, idFile);
 			if (line && file) {
-				char *quote = strstr(file, "'");
-				size_t length = quote - (file + lenFile + 1);
+				char *fileStart = file + lenFile + 1;
+				char *quote = strstr(fileStart, "'");
+				size_t length = quote - fileStart;
 				if (quote && length > 0) {
-					strncpy(sourcePath, file + lenFile + 1, length);
+					strncpy(sourcePath, fileStart, length);
 					sourcePath[length] = '\0';
 				}
 				line += lenLine;
@@ -1556,7 +1586,7 @@ int DecodeMessage(char *cdoc, char *sourcePath, int format, int &column) {
 			/* Java runtime stack trace
 				\tat <methodname>(<filename>:<line>)
 				 */
-			char *startPath = strchr(cdoc, '(') + 1;
+			char *startPath = strrchr(cdoc, '(') + 1;
 			char *endPath = strchr(startPath, ':');
 			int length = endPath - startPath;
 			if (length > 0) {
@@ -1565,9 +1595,24 @@ int DecodeMessage(char *cdoc, char *sourcePath, int format, int &column) {
 				int sourceNumber = atoi(endPath+1) - 1;
 				return sourceNumber;
 			}
+			break;
+		}
+
+	case SCE_ERR_DIFF_MESSAGE: {
+			// Diff file header, either +++ <filename>\t or --- <filename>\t
+			// Often followed by a position line @@ <linenumber>
+			char *startPath = cdoc + 4;
+			char *endPath = strchr(startPath, '\t');
+			if (endPath) {
+				int length = endPath - startPath;
+				strncpy(sourcePath, startPath, length);
+				sourcePath[length] = 0;
+				return 0;
+			}
+			break;
 		}
 	}	// switch
-	return - 1;
+	return -1;
 }
 
 void SciTEBase::GoMessage(int dir) {
@@ -1589,8 +1634,10 @@ void SciTEBase::GoMessage(int dir) {
 		//Platform::DebugPrintf("GOMessage %d %d %d of %d linestart = %d\n", selStart, curLine, lookLine, maxLine, startPosLine);
 		char style = acc.StyleAt(startPosLine);
 		if (style != SCE_ERR_DEFAULT &&
-		        style != SCE_ERR_CMD &&
-		        (style < SCE_ERR_DIFF_CHANGED || style > SCE_ERR_DIFF_MESSAGE)) {
+			style != SCE_ERR_CMD &&
+			style != SCE_ERR_DIFF_ADDITION &&
+			style != SCE_ERR_DIFF_CHANGED &&
+			style != SCE_ERR_DIFF_DELETION) {
 			//Platform::DebugPrintf("Marker to %d\n", lookLine);
 			SendOutput(SCI_MARKERDELETEALL, static_cast<uptr_t>(-1));
 			SendOutput(SCI_MARKERDEFINE, 0, SC_MARK_SMALLRECT);
@@ -1605,22 +1652,31 @@ void SciTEBase::GoMessage(int dir) {
 				return;
 			GetRange(wOutput, startPosLine, startPosLine + lineLength, cdoc);
 			SString message(cdoc);
-			char sourcePath[MAX_PATH];
+			char source[MAX_PATH];
 			int column;
-			int sourceLine = DecodeMessage(cdoc, sourcePath, style, column);
-			//Platform::DebugPrintf("<%s> %d %d\n",sourcePath, sourceLine, lookLine);
+			int sourceLine = DecodeMessage(cdoc, source, style, column);
+			//Platform::DebugPrintf("<%s> %d %d\n",source, sourceLine, lookLine);
 			if (sourceLine >= 0) {
-				if (0 != strcmp(sourcePath, fileName)) {
-					char messagePath[MAX_PATH];
+				FilePath sourcePath(source);
+				if (!filePath.Name().SameNameAs(sourcePath)) {
+					FilePath messagePath;
 					bool bExists = false;
-					if (Exists(dirNameAtExecute.c_str(), sourcePath, messagePath)) {
+					if (Exists(dirNameAtExecute.AsInternal(), source, &messagePath)) {
 						bExists = true;
-					} else if (Exists(dirNameForExecute.c_str(), sourcePath, messagePath)) {
+					} else if (Exists(dirNameForExecute.AsInternal(), source, &messagePath)) {
 						bExists = true;
-					} else if (Exists(dirName, sourcePath, messagePath)) {
+					} else if (Exists(filePath.Directory().AsInternal(), source, &messagePath)) {
 						bExists = true;
-					} else if (Exists(NULL, sourcePath, messagePath)) {
+					} else if (Exists(NULL, source, &messagePath)) {
 						bExists = true;
+					} else {
+						// Look through buffers for name match
+						for (int i = buffers.length - 1; i >= 0; i--) {
+							if (sourcePath.Name().SameNameAs(buffers.buffers[i].Name())) {
+								messagePath = buffers.buffers[i];
+								bExists = true;
+							}
+						}
 					}
 					if (bExists) {
 						if (!Open(messagePath)) {
@@ -1661,7 +1717,7 @@ void SciTEBase::GoMessage(int dir) {
 				int endSourceline= SendEditor(SCI_POSITIONFROMLINE, sourceLine+1, 0);
 				if (column >= 0) {
 					// Get the position in line according to current tab setting
-					startSourceLine = SendEditor(SCI_FINDCOLUMN, sourceLine, column); 
+					startSourceLine = SendEditor(SCI_FINDCOLUMN, sourceLine, column);
 				}
 				EnsureRangeVisible(startSourceLine, startSourceLine);
 				if (props.GetInt("error.select.line")==1) {
