@@ -10,17 +10,10 @@
 #include <ctype.h>
 #include <stdio.h>
 #include <fcntl.h>
-#include <sys/stat.h>
 #include <time.h>
 
 #include "Platform.h"
 
-#if PLAT_FOX
-
-#include <unistd.h>
-
-#endif
-
 #if PLAT_GTK
 
 #include <unistd.h>
@@ -59,254 +52,82 @@
 #include "Scintilla.h"
 #include "Extender.h"
 #include "Utf8_16.h"
+#include "FilePath.h"
+#include "PropSetFile.h"
 #include "SciTEBase.h"
 
 #ifdef unix
-const char pathSepString[] = "/";
-const char pathSepChar = '/';
-const char listSepString[] = ":";
-const char configFileVisibilityString[] = ".";
 const char propUserFileName[] = ".SciTEUser.properties";
-const char fileRead[] = "rb";
-const char fileWrite[] = "wb";
-#endif
-#ifdef __vms
-const char pathSepString[] = "/";
-const char pathSepChar = '/';
-const char listSepString[] = ":";
-const char configFileVisibilityString[] = "";
+#else
+// Windows or VMS
 const char propUserFileName[] = "SciTEUser.properties";
-const char fileRead[] = "r";
-const char fileWrite[] = "w";
-#endif
-#ifdef WIN32
-// Windows
-const char pathSepString[] = "\\";
-const char pathSepChar = '\\';
-const char listSepString[] = ";";
-const char configFileVisibilityString[] = "";
-const char propUserFileName[] = "SciTEUser.properties";
-const char fileRead[] = "rb";
-const char fileWrite[] = "wb";
 #endif
 const char propGlobalFileName[] = "SciTEGlobal.properties";
 const char propAbbrevFileName[] = "abbrev.properties";
 
 #define PROPERTIES_EXTENSION	".properties"
 
-static bool IsPropertiesFile(char *filename) {
-	size_t nameLen = strlen(filename);
-	size_t propLen = strlen(PROPERTIES_EXTENSION);
-	if (nameLen <= propLen)
-		return false;
-	if (EqualCaseInsensitive(filename + nameLen - propLen, PROPERTIES_EXTENSION))
+static bool IsPropertiesFile(const FilePath &filename) {
+	FilePath ext = filename.Extension();
+	if (EqualCaseInsensitive(ext.AsInternal(), PROPERTIES_EXTENSION + 1))
 		return true;
 	return false;
 }
 
-bool SciTEBase::IsAbsolutePath(const char *path) {
-	if (!path || *path == '\0')
-		return false;
-#ifdef unix
-
-	if (path[0] == '/')
-		return true;
-#endif
-#ifdef __vms
-
-	if (path[0] == '/')
-		return true;
-#endif
-#ifdef WIN32
-
-	if (path[0] == '\\' || path[1] == ':')	// UNC path or drive separator
-		return true;
-#endif
-
-	return false;
-}
-
-void SciTEBase::FixFilePath() {
-	// Only used on Windows to fix the case of file names
-}
-
-#ifdef __vms
-const char *VMSToUnixStyle(const char *fileName) {
-	// possible formats:
-	// o disk:[dir.dir]file.type
-	// o logical:file.type
-	// o [dir.dir]file.type
-	// o file.type
-	// o /disk//dir/dir/file.type
-	// o /disk/dir/dir/file.type
-
-	static char unixStyleFileName[MAX_PATH + 20];
-
-	if (strchr(fileName, ':') == NULL && strchr(fileName, '[') == NULL) {
-		// o file.type
-		// o /disk//dir/dir/file.type
-		// o /disk/dir/dir/file.type
-		if (strstr (fileName, "//") == NULL) {
-			return fileName;
-		}
-		strcpy(unixStyleFileName, fileName);
-		char *p;
-		while ((p = strstr (unixStyleFileName, "//")) != NULL) {
-			strcpy (p, p + 1);
-		}
-		return unixStyleFileName;
-	}
-
-	// o disk:[dir.dir]file.type
-	// o logical:file.type
-	// o [dir.dir]file.type
-
-	if (fileName [0] == '/') {
-		strcpy(unixStyleFileName, fileName);
-	} else {
-		unixStyleFileName [0] = '/';
-		strcpy(unixStyleFileName + 1, fileName);
-		char *p = strstr(unixStyleFileName, ":[");
-		if (p == NULL) {
-			// o logical:file.type
-			p = strchr(unixStyleFileName, ':');
-			*p = '/';
-		} else {
-			*p = '/';
-			strcpy(p + 1, p + 2);
-			char *end = strchr(unixStyleFileName, ']');
-			if (*end != NULL) {
-				*end = '/';
-			}
-			while (p = strchr(unixStyleFileName, '.'), p != NULL && p < end) {
-				*p = '/';
-			}
-		}
-	}
-	return unixStyleFileName;
-} // VMSToUnixStyle
-#endif
-
-void SciTEBase::SetFileName(const char *openName, bool fixCase) {
-	if (openName[0] == '\"') {
+void SciTEBase::SetFileName(FilePath openName, bool fixCase) {
+	if (openName.AsInternal()[0] == '\"') {
 		// openName is surrounded by double quotes
 		char pathCopy[MAX_PATH + 1];
-		pathCopy[0] = '\0';
-		strncpy(pathCopy, openName + 1, MAX_PATH);
+		strncpy(pathCopy, openName.AsInternal() + 1, MAX_PATH);
 		pathCopy[MAX_PATH] = '\0';
 		if (pathCopy[strlen(pathCopy) - 1] == '\"') {
 			pathCopy[strlen(pathCopy) - 1] = '\0';
 		}
-		AbsolutePath(fullPath, pathCopy, MAX_PATH);
-	} else if (openName[0]) {
-		AbsolutePath(fullPath, openName, MAX_PATH);
+		filePath.Set(pathCopy);
 	} else {
-		fullPath[0] = '\0';
+		filePath.Set(openName);
 	}
 
 	// Break fullPath into directory and file name using working directory for relative paths
-	dirName[0] = '\0';
-	char *cpDirEnd = strrchr(fullPath, pathSepChar);
-	if (IsAbsolutePath(fullPath)) {
-		// Absolute path
-		strcpy(fileName, cpDirEnd + 1);
-		strcpy(dirName, fullPath);
-		dirName[cpDirEnd - fullPath] = '\0';
-		//Platform::DebugPrintf("SetFileName: <%s> <%s>\n", fileName, dirName);
-	} else {
+	if (!filePath.IsAbsolute()) {
 		// Relative path. Since we ran AbsolutePath, we probably are here because fullPath is empty.
-		GetDocumentDirectory(dirName, sizeof(dirName));
-		//Platform::DebugPrintf("Working directory: <%s>\n", dirName);
-		if (cpDirEnd) {
-			// directories and file name
-			strcpy(fileName, cpDirEnd + 1);
-			*cpDirEnd = '\0';
-			strncat(dirName, pathSepString, sizeof(dirName));
-			strncat(dirName, fullPath, sizeof(dirName));
-		} else {
-			// Just a file name
-			strcpy(fileName, fullPath);
-		}
+		filePath.SetDirectory(filePath.Directory());
 	}
 
-	// Rebuild fullPath from directory and name
-	strcpy(fullPath, dirName);
-	strcat(fullPath, pathSepString);
-	strcat(fullPath, fileName);
-	//Platform::DebugPrintf("Path: <%s>\n", fullPath);
-
 	if (fixCase) {
-		FixFilePath();
-	}
-
-	char fileBase[MAX_PATH];
-	strcpy(fileBase, fileName);
-	char *extension = strrchr(fileBase, '.');
-	if (extension) {
-		*extension = '\0';
-		strcpy(fileExt, extension + 1);
-	} else { // No extension
-		fileExt[0] = '\0';
+		filePath.FixName();
 	}
 
 	ReadLocalPropFile();
 
-	props.Set("FilePath", fullPath);
-	props.Set("FileDir", dirName);
-	props.Set("FileName", fileBase);
-	props.Set("FileExt", fileExt);
-	props.Set("FileNameExt", fileName);
+	props.Set("FilePath", filePath.AsFileSystem());
+	props.Set("FileDir", filePath.Directory().AsFileSystem());
+	props.Set("FileName", filePath.BaseName().AsFileSystem());
+	props.Set("FileExt", filePath.Extension().AsFileSystem());
+	props.Set("FileNameExt", FileNameExt().AsFileSystem());
 
 	SetWindowName();
 	if (buffers.buffers)
-		buffers.buffers[buffers.Current()].Set(fullPath);
+		buffers.buffers[buffers.Current()].Set(filePath);
 	BuffersMenu();
 }
 
-bool SciTEBase::Exists(const char *dir, const char *path, char *testPath) {
-	char copyPath[MAX_PATH];
-	if (IsAbsolutePath(path) || !dir) {
-		strcpy(copyPath, path);
-	} else if (dir) {
-		if ((strlen(dir) + strlen(pathSepString) + strlen(path) + 1) > MAX_PATH)
-			return false;
-		strcpy(copyPath, dir);
-		strcat(copyPath, pathSepString);
-		strcat(copyPath, path);
+// See if path exists.
+// If path is not absolute, it is combined with dir.
+// If resultPath is not NULL, it receives the absolute path if it exists.
+bool SciTEBase::Exists(const char *dir, const char *path, FilePath *resultPath) {
+	FilePath copy(path);
+	if (!copy.IsAbsolute() && dir) {
+		copy.SetDirectory(dir);
 	}
-	if (!copyPath[0])
+	if (!copy.Exists())
 		return false;
-	FILE *fp = fopen(copyPath, fileRead);
-	if (!fp)
-		return false;
-	fclose(fp);
-	if (testPath)
-		AbsolutePath(testPath, copyPath, MAX_PATH);
+	if (resultPath) {
+		resultPath->Set(copy.AbsolutePath());
+	}
 	return true;
 }
 
-bool BuildPath(char *path, const char *dir, const char *fileName,
-               unsigned int lenPath) {
-	*path = '\0';
-	if ((strlen(dir) + strlen(pathSepString) + strlen(fileName)) < lenPath) {
-		strcpy(path, dir);
-		strcat(path, pathSepString);
-		strcat(path, fileName);
-		return true;
-	}
-	return false;
-}
-
-time_t GetModTime(const char *fullPath) {
-	if (IsUntitledFileName(fullPath))
-		return 0;
-	struct stat statusFile;
-	if (stat(fullPath, &statusFile) != -1)
-		return statusFile.st_mtime;
-	else
-		return 0;
-}
-
 void SciTEBase::CountLineEnds(int &linesCR, int &linesLF, int &linesCRLF) {
 	linesCR = 0;
 	linesLF = 0;
@@ -387,6 +208,16 @@ static UniMode CookieValue(const SString &s) {
 	return uni8Bit;
 }
 
+static UniMode CodingCookieValue(const char *buf, size_t length) {
+	SString l1 = ExtractLine(buf, length);
+	UniMode unicodeMode = CookieValue(l1);
+	if (unicodeMode == uni8Bit) {
+		SString l2 = ExtractLine(buf+l1.length(), length-l1.length());
+		unicodeMode = CookieValue(l2);
+	}
+	return unicodeMode;
+}
+
 void SciTEBase::DiscoverEOLSetting() {
 	int linesCR;
 	int linesLF;
@@ -401,6 +232,43 @@ void SciTEBase::DiscoverEOLSetting() {
 		SendEditor(SCI_SETEOLMODE, SC_EOL_CRLF);
 }
 
+// Look inside the first line for a #! clue regarding the language
+SString SciTEBase::DiscoverLanguage(const char *buf, size_t length) {
+	SString languageOverride = "";
+	SString l1 = ExtractLine(buf, length);
+	if (l1.startswith("<?xml")) {
+		languageOverride = "xml";
+	} else if (l1.startswith("#!")) {
+		l1 = l1.substr(2);
+		l1.substitute('\\', ' ');
+		l1.substitute('/', ' ');
+		l1.substitute("\t", " ");
+		l1.substitute("  ", " ");
+		l1.substitute("  ", " ");
+		l1.substitute("  ", " ");
+		l1.remove("\r");
+		l1.remove("\n");
+		if (l1.startswith(" ")) {
+			l1 = l1.substr(1);
+		}
+		l1.substitute(' ', '\0');
+		const char *word = l1.c_str();
+		while (*word) {
+			SString propShBang("shbang.");
+			propShBang.append(word);
+			SString langShBang = props.GetExpanded(propShBang.c_str());
+			if (langShBang.length()) {
+				languageOverride = langShBang;
+			}
+			word += strlen(word) + 1;
+		}
+	}
+	if (languageOverride.length()) {
+		languageOverride.insert(0, "x.");
+	}
+	return languageOverride;
+}
+
 void SciTEBase::DiscoverIndentSetting() {
 	int lengthDoc = LengthDocument();
 	WindowAccessor acc(wEditor.GetID(), props);
@@ -457,34 +325,41 @@ void SciTEBase::DiscoverIndentSetting() {
 void SciTEBase::OpenFile(int fileSize, bool suppressMessage) {
 	Utf8_16_Read convert;
 
-	FILE *fp = fopen(fullPath, fileRead);
+	FILE *fp = filePath.Open(fileRead);
 	if (fp) {
-		fileModTime = GetModTime(fullPath);
-		fileModLastAsk = fileModTime;
+		CurrentBuffer()->SetTimeFromFile();
 		SendEditor(SCI_BEGINUNDOACTION);	// Group together clear and insert
 		SendEditor(SCI_CLEARALL);
 		char data[blockSize];
 		size_t lenFile = fread(data, 1, sizeof(data), fp);
-		SString l1 = ExtractLine(data, lenFile);
-		SString l2 = ExtractLine(data+l1.length(), lenFile-l1.length());
+		UniMode codingCookie = CodingCookieValue(data, lenFile);
 		SendEditor(SCI_ALLOCATE, fileSize + 1000);
+		SString languageOverride;
+		bool firstBlock = true;
 		while (lenFile > 0) {
 			lenFile = convert.convert(data, lenFile);
-			SendEditorString(SCI_ADDTEXT, lenFile, convert.getNewBuf());
+			char *dataBlock = convert.getNewBuf();
+			if ((firstBlock) && (language == "")) {
+				languageOverride = DiscoverLanguage(dataBlock, lenFile);
+			}
+			firstBlock = false;
+			SendEditorString(SCI_ADDTEXT, lenFile, dataBlock);
 			lenFile = fread(data, 1, sizeof(data), fp);
 		}
 		fclose(fp);
 		SendEditor(SCI_ENDUNDOACTION);
-		unicodeMode = static_cast<UniMode>(
+		if (languageOverride.length()) {
+			CurrentBuffer()->overrideExtension = languageOverride;
+			ReadProperties();
+			SetIndentSettings();
+		}
+		CurrentBuffer()->unicodeMode = static_cast<UniMode>(
 			static_cast<int>(convert.getEncoding()));
 		// Check the first two lines for coding cookies
-		if (unicodeMode == uni8Bit) {
-			unicodeMode = CookieValue(l1);
-			if (unicodeMode == uni8Bit) {
-				unicodeMode = CookieValue(l2);
-			}
+		if (CurrentBuffer()->unicodeMode == uni8Bit) {
+			CurrentBuffer()->unicodeMode = codingCookie;
 		}
-		if (unicodeMode != uni8Bit) {
+		if (CurrentBuffer()->unicodeMode != uni8Bit) {
 			// Override the code page if Unicode
 			codePage = SC_CP_UTF8;
 		} else {
@@ -501,7 +376,7 @@ void SciTEBase::OpenFile(int fileSize, bool suppressMessage) {
 		}
 
 	} else if (!suppressMessage) {
-		SString msg = LocaliseMessage("Could not open file '^0'.", fullPath);
+		SString msg = LocaliseMessage("Could not open file '^0'.", filePath.AsFileSystem());
 		WindowMessageBox(wSciTE, msg, MB_OK | MB_ICONWARNING);
 	}
 	SendEditor(SCI_SETUNDOCOLLECTION, 1);
@@ -521,36 +396,13 @@ bool SciTEBase::PreOpenCheck(const char *) {
 	return false;
 }
 
-static int GetFileLength(const char *filePath) {
-	int size = -1;
-	if (filePath && *filePath) {
-		FILE *fp = fopen(filePath, fileRead);
-		if (fp) {
-			fseek(fp, 0, SEEK_END);
-			size = ftell(fp);
-			fseek(fp, 0, SEEK_SET);
-			fclose(fp);
-		}
-	}
-	return size;
-}
-
-bool SciTEBase::Open(const char *file, OpenFlags of) {
+bool SciTEBase::Open(FilePath file, OpenFlags of) {
 	InitialiseBuffers();
 
-	if (!file) {
-		SString msg = LocaliseMessage("Bad file.");
-		WindowMessageBox(wSciTE, msg, MB_OK | MB_ICONWARNING);
-		return false;
-	}
 #ifdef __vms
-	static char fixedFileName[MAX_PATH];
-	strcpy(fixedFileName, VMSToUnixStyle(file));
-	file = fixedFileName;
+	file = file.VMSToUnixStyle();
 #endif
-
-	char absPath[MAX_PATH];
-	AbsolutePath(absPath, file, MAX_PATH);
+	FilePath absPath = file.AbsolutePath();
 	int index = buffers.GetDocumentByName(absPath);
 	if (index >= 0) {
 		SetDocumentAt(index);
@@ -564,17 +416,16 @@ bool SciTEBase::Open(const char *file, OpenFlags of) {
 		return false;
 	}
 
-	int size = 0;
-	if (absPath[0] != '\0') {
+	int size = absPath.GetFileLength();
+	if (size > 0) {
 		// Real file, not empty buffer
-		size = GetFileLength(absPath);
 		int maxSize = props.GetInt("max.file.size");
 		if (maxSize > 0 && size > maxSize) {
 			SString sSize(size), sMaxSize(maxSize);
 			SString msg = LocaliseMessage("File '^0' is ^1 bytes long,\n"
 						      "larger than the ^2 bytes limit set in the properties.\n"
 						      "Do you still want to open it?",
-						      absPath, sSize.c_str(), sMaxSize.c_str());
+						      absPath.AsFileSystem(), sSize.c_str(), sMaxSize.c_str());
 			int answer = WindowMessageBox(wSciTE, msg, MB_YESNO | MB_ICONWARNING);
 			if (answer != IDYES) {
 				return false;
@@ -583,7 +434,7 @@ bool SciTEBase::Open(const char *file, OpenFlags of) {
 	}
 
 	if (buffers.size == buffers.length) {
-		AddFileToStack(fullPath, GetSelection(), GetCurrentScrollPosition());
+		AddFileToStack(filePath, GetSelection(), GetCurrentScrollPosition());
 		ClearDocument();
 		if (extender)
 			extender->InitBuffer(buffers.Current());
@@ -595,13 +446,13 @@ bool SciTEBase::Open(const char *file, OpenFlags of) {
 
 	//Platform::DebugPrintf("Opening %s\n", file);
 	SetFileName(absPath);
-	overrideExtension = "";
+	CurrentBuffer()->overrideExtension = "";
 	ReadProperties();
 	SetIndentSettings();
 	UpdateBuffersCurrent();
 	SizeSubWindows();
 
-	if (fileName[0]) {
+	if (!filePath.IsUntitled()) {
 		SendEditor(SCI_SETREADONLY, 0);
 		SendEditor(SCI_CANCEL);
 		if (of & ofPreserveUndo) {
@@ -620,7 +471,7 @@ bool SciTEBase::Open(const char *file, OpenFlags of) {
 		isReadOnly = props.GetInt("read.only");
 		SendEditor(SCI_SETREADONLY, isReadOnly);
 	}
-	RemoveFileFromStack(fullPath);
+	RemoveFileFromStack(filePath);
 	DeleteFileStackMenu();
 	SetFileStackMenu();
 	SetWindowName();
@@ -628,7 +479,7 @@ bool SciTEBase::Open(const char *file, OpenFlags of) {
 		SetLineNumberWidth();
 	UpdateStatusBar(true);
 	if (extender)
-		extender->OnOpen(fullPath);
+		extender->OnOpen(filePath.AsFileSystem());
 	return true;
 }
 
@@ -649,13 +500,13 @@ bool SciTEBase::OpenSelected() {
 	SString openSuffix = props.GetNewExpand("open.suffix.", fileNameForExtension.c_str());
 	strcat(selectedFilename, openSuffix.c_str());
 
-	if (EqualCaseInsensitive(selectedFilename, fileName) || EqualCaseInsensitive(selectedFilename, fullPath)) {
+	if (EqualCaseInsensitive(selectedFilename, FileNameExt().AsInternal()) || EqualCaseInsensitive(selectedFilename, filePath.AsInternal())) {
 		WarnUser(warnWrongFile);
 		return true;	// Do not open if it is the current file!
 	}
 
 	cTag[0] = '\0';
-	if (IsPropertiesFile(fileName) &&
+	if (IsPropertiesFile(filePath) &&
 			strchr(selectedFilename, '.') == 0 &&
 	        strlen(selectedFilename) + strlen(PROPERTIES_EXTENSION) < MAX_PATH) {
 		// We are in a properties file and try to open a file without extension,
@@ -701,14 +552,13 @@ bool SciTEBase::OpenSelected() {
 		}
 	}
 
-	char path[MAX_PATH];
-	*path = '\0';
+	FilePath path;
 	// Don't load the path of the current file if the selected
 	// filename is an absolute pathname
-	if (!IsAbsolutePath(selectedFilename)) {
-		GetDocumentDirectory(path, sizeof(path));
+	if (!FilePath(selectedFilename).IsAbsolute()) {
+		path = filePath.Directory();
 		// If not there, look in openpath
-		if (!Exists(path, selectedFilename, NULL)) {
+		if (!Exists(path.AsInternal(), selectedFilename, NULL)) {
 			SString openPath = props.GetNewExpand(
 				"openpath.", fileNameForExtension.c_str());
 			while (openPath.length()) {
@@ -721,15 +571,15 @@ bool SciTEBase::OpenSelected() {
 					openPath.clear();
 				}
 				if (Exists(tryPath.c_str(), selectedFilename, NULL)) {
-					strncpy(path, tryPath.c_str(), sizeof(path)-1);
-					path[sizeof(path)-1] = '\0';
+					path.Set(tryPath.c_str());
 					break;
 				}
 			}
 		}
 	}
-	if (Exists(path, selectedFilename, path)) {
-		if (Open(path)) {
+	FilePath pathReturned;
+	if (Exists(path.AsInternal(), selectedFilename, &pathReturned)) {
+		if (Open(pathReturned)) {
 			if (lineNumber > 0) {
 				SendEditor(SCI_GOTOLINE, lineNumber - 1);
 			} else if (cTag[0] != '\0') {
@@ -750,41 +600,39 @@ bool SciTEBase::OpenSelected() {
 
 void SciTEBase::Revert() {
 	RecentFile rf = GetFilePosition();
-	OpenFile(GetFileLength(fullPath), false);
+	OpenFile(filePath.GetFileLength(), false);
 	DisplayAround(rf);
 }
 
 void SciTEBase::CheckReload() {
 	if (props.GetInt("load.on.activate")) {
 		// Make a copy of fullPath as otherwise it gets aliased in Open
-		char fullPathToCheck[MAX_PATH];
-		strcpy(fullPathToCheck, fullPath);
-		time_t newModTime = GetModTime(fullPathToCheck);
-		//Platform::DebugPrintf("Times are %d %d\n", fileModTime, newModTime);
-		if (newModTime > fileModTime) {
+		time_t newModTime = filePath.ModifiedTime();
+		//Platform::DebugPrintf("Times are %d %d\n", CurrentBuffer()->fileModTime, newModTime);
+		if (newModTime > CurrentBuffer()->fileModTime) {
 			RecentFile rf = GetFilePosition();
 			OpenFlags of = props.GetInt("reload.preserves.undo") ? ofPreserveUndo : ofNone;
-			if (isDirty || props.GetInt("are.you.sure.on.reload") != 0) {
-				if ((0 == dialogsOnScreen) && (newModTime != fileModLastAsk)) {
+			if (CurrentBuffer()->isDirty || props.GetInt("are.you.sure.on.reload") != 0) {
+				if ((0 == dialogsOnScreen) && (newModTime != CurrentBuffer()->fileModLastAsk)) {
 					SString msg;
-					if (isDirty) {
+					if (CurrentBuffer()->isDirty) {
 						msg = LocaliseMessage(
 							  "The file '^0' has been modified. Should it be reloaded?",
-							  fullPathToCheck);
+							  filePath.AsFileSystem());
 					} else {
 						msg = LocaliseMessage(
 							  "The file '^0' has been modified outside SciTE. Should it be reloaded?",
-							  fileName);
+							  FileNameExt().AsFileSystem());
 					}
 					int decision = WindowMessageBox(wSciTE, msg, MB_YESNO);
 					if (decision == IDYES) {
-						Open(fullPathToCheck, static_cast<OpenFlags>(of | ofForceLoad));
+						Open(filePath, static_cast<OpenFlags>(of | ofForceLoad));
 						DisplayAround(rf);
 					}
-					fileModLastAsk = newModTime;
+					CurrentBuffer()->fileModLastAsk = newModTime;
 				}
 			} else {
-				Open(fullPathToCheck, static_cast<OpenFlags>(of | ofForceLoad));
+				Open(filePath, static_cast<OpenFlags>(of | ofForceLoad));
 				DisplayAround(rf);
 			}
 		}
@@ -801,14 +649,35 @@ void SciTEBase::Activate(bool activeApp) {
 	}
 }
 
+FilePath SciTEBase::SaveName(const char *ext) {
+	SString savePath = filePath.AsInternal();
+	if (ext) {
+		int dot = savePath.length()-1;
+		while ((dot >= 0) && (savePath[dot] != '.')) {
+			dot--;
+		}
+		if (dot >= 0) {
+			int keepExt = props.GetInt("export.keep.ext");
+			if (keepExt == 0) {
+				savePath.remove(dot, 0);
+			} else if (keepExt == 2) {
+				savePath.change(dot, '_');
+			}
+		}
+		savePath.append(ext);
+	}
+	//~ fprintf(stderr, "SaveName <%s> <%s> <%s>\n", filePath.AsInternal(), savePath.c_str(), ext);
+	return FilePath(savePath.c_str());
+}
+
 int SciTEBase::SaveIfUnsure(bool forceQuestion) {
-	if ((isDirty) && (LengthDocument() || fileName[0] || forceQuestion)) {
+	if ((CurrentBuffer()->isDirty) && (LengthDocument() || !filePath.IsUntitled() || forceQuestion)) {
 		if (props.GetInt("are.you.sure", 1) ||
-		        IsUntitledFileName(fullPath) ||
+		        filePath.IsUntitled() ||
 		        forceQuestion) {
 			SString msg;
-			if (fileName[0]) {
-				msg = LocaliseMessage("Save changes to '^0'?", fullPath);
+			if (!filePath.IsUntitled()) {
+				msg = LocaliseMessage("Save changes to '^0'?", filePath.AsFileSystem());
 			} else {
 				msg = LocaliseMessage("Save changes to (Untitled)?");
 			}
@@ -830,15 +699,14 @@ int SciTEBase::SaveIfUnsureAll(bool forceQuestion) {
 	if (SaveAllBuffers(forceQuestion) == IDCANCEL) {
 		return IDCANCEL;
 	}
-	if (props.GetInt("save.recent", 0)) {
+	if (props.GetInt("save.recent")) {
 		for (int i = 0; i < buffers.length; ++i) {
 			Buffer buff = buffers.buffers[i];
-			AddFileToStack(buff.FullPath(),
-			               buff.selection, buff.scrollPosition);
+			AddFileToStack(buff, buff.selection, buff.scrollPosition);
 		}
 		SaveRecentStack();
 	}
-	if (props.GetInt("buffers") && props.GetInt("save.session", 0))
+	if (props.GetInt("buffers") && props.GetInt("save.session"))
 		SaveSession("");
 
 	// Definitely going to exit now, so delete all documents
@@ -854,8 +722,11 @@ int SciTEBase::SaveIfUnsureAll(bool forceQuestion) {
 }
 
 int SciTEBase::SaveIfUnsureForBuilt() {
-	if (isDirty) {
-		if (props.GetInt("are.you.sure.for.build", 0))
+	if (props.GetInt("save.all.for.build")) {
+		return SaveAllBuffers(false, !props.GetInt("are.you.sure.for.build"));
+	}
+	if (CurrentBuffer()->isDirty) {
+		if (props.GetInt("are.you.sure.for.build"))
 			return SaveIfUnsure(true);
 
 		Save();
@@ -906,7 +777,7 @@ void SciTEBase::EnsureFinalNewLine() {
 /**
  * Writes the buffer to the given filename.
  */
-bool SciTEBase::SaveBuffer(const char *saveName) {
+bool SciTEBase::SaveBuffer(FilePath saveName) {
 	bool retVal = false;
 	// Perform clean ups on text before saving
 	SendEditor(SCI_BEGINUNDOACTION);
@@ -918,18 +789,19 @@ bool SciTEBase::SaveBuffer(const char *saveName) {
 		SendEditor(SCI_CONVERTEOLS, SendEditor(SCI_GETEOLMODE));
 
 	if (extender)
-		extender->OnBeforeSave(saveName);
+		extender->OnBeforeSave(saveName.AsFileSystem());
 
 	SendEditor(SCI_ENDUNDOACTION);
 
 	Utf8_16_Write convert;
-	if (unicodeMode != uniCookie) {	// Save file with cookie without BOM.
+	if (CurrentBuffer()->unicodeMode != uniCookie) {	// Save file with cookie without BOM.
 		convert.setEncoding(static_cast<Utf8_16::encodingType>(
-			static_cast<int>(unicodeMode)));
+			static_cast<int>(CurrentBuffer()->unicodeMode)));
 	}
 
-	FILE *fp = convert.fopen(saveName, fileWrite);
+	FILE *fp = saveName.Open(fileWrite);
 	if (fp) {
+		convert.setfile(fp);
 		char data[blockSize + 1];
 		int lengthDoc = LengthDocument();
 		retVal = true;
@@ -948,7 +820,7 @@ bool SciTEBase::SaveBuffer(const char *saveName) {
 		convert.fclose();
 
 		if (extender)
-			extender->OnSave(saveName);
+			extender->OnSave(saveName.AsFileSystem());
 	}
 	UpdateStatusBar(true);
 	return retVal;
@@ -967,44 +839,316 @@ void SciTEBase::ReloadProperties() {
 
 // Returns false if cancelled or failed to save
 bool SciTEBase::Save() {
-	if (fileName[0]) {
+	if (!filePath.IsUntitled()) {
 		if (props.GetInt("save.deletes.first")) {
-			unlink(fullPath);
+			filePath.Remove();
 		}
 
-		if (SaveBuffer(fullPath)) {
-			fileModTime = GetModTime(fullPath);
+		if (SaveBuffer(filePath)) {
+			CurrentBuffer()->SetTimeFromFile();
 			SendEditor(SCI_SETSAVEPOINT);
-			if (IsPropertiesFile(fileName)) {
+			if (IsPropertiesFile(filePath)) {
 				ReloadProperties();
 			}
 		} else {
 			SString msg = LocaliseMessage(
-				"Could not save file '^0'. Save under a different name?", fullPath);
+				"Could not save file '^0'. Save under a different name?", filePath.AsFileSystem());
 			int decision = WindowMessageBox(wSciTE, msg, MB_YESNO | MB_ICONWARNING);
 			if (decision == IDYES) {
-				return SaveAs();
+				return SaveAsDialog();
 			}
 			return false;
 		}
 		return true;
-	} else {
-		return SaveAs();
-	}
-}
-
-bool SciTEBase::SaveAs(const char *file) {
-	if (file && *file) {
-		SetFileName(file);
-		Save();
-		ReadProperties();
-		SendEditor(SCI_CLEARDOCUMENTSTYLE);
-		SendEditor(SCI_COLOURISE, 0, -1);
-		Redraw();
-		SetWindowName();
-		BuffersMenu();
-		return true;
 	} else {
 		return SaveAsDialog();
 	}
 }
+
+void SciTEBase::SaveAs(const char *file) {
+	SetFileName(file);
+	Save();
+	ReadProperties();
+	SendEditor(SCI_CLEARDOCUMENTSTYLE);
+	SendEditor(SCI_COLOURISE, 0, -1);
+	Redraw();
+	SetWindowName();
+	BuffersMenu();
+}
+
+bool SciTEBase::IsStdinBlocked() {
+	return false; /* always default to blocked */
+}
+
+void SciTEBase::OpenFromStdin(bool UseOutputPane) {
+	Utf8_16_Read convert;
+	char data[blockSize];
+
+	/* if stdin is blocked, do not execute this method */
+	if (IsStdinBlocked())
+		return;
+
+	Open("");
+	if (UseOutputPane) {
+		SendOutput(SCI_CLEARALL);
+	} else {
+		SendEditor(SCI_BEGINUNDOACTION);	// Group together clear and insert
+		SendEditor(SCI_CLEARALL);
+	}
+	size_t lenFile = fread(data, 1, sizeof(data), stdin);
+	UniMode codingCookie = CodingCookieValue(data, lenFile);
+	while (lenFile > 0) {
+		lenFile = convert.convert(data, lenFile);
+		if (UseOutputPane) {
+			SendOutputString(SCI_ADDTEXT, lenFile, convert.getNewBuf());
+		} else {
+			SendEditorString(SCI_ADDTEXT, lenFile, convert.getNewBuf());
+		}
+		lenFile = fread(data, 1, sizeof(data), stdin);
+	}
+	if (UseOutputPane) {
+		if (props.GetInt("split.vertical") == 0) {
+			heightOutput = 2000;
+		} else {
+			heightOutput = 500;
+		}
+		SizeSubWindows();
+	} else {
+		SendEditor(SCI_ENDUNDOACTION);
+	}
+	CurrentBuffer()->unicodeMode = static_cast<UniMode>(
+		static_cast<int>(convert.getEncoding()));
+	// Check the first two lines for coding cookies
+	if (CurrentBuffer()->unicodeMode == uni8Bit) {
+		CurrentBuffer()->unicodeMode = codingCookie;
+	}
+	if (CurrentBuffer()->unicodeMode != uni8Bit) {
+		// Override the code page if Unicode
+		codePage = SC_CP_UTF8;
+	} else {
+		codePage = props.GetInt("code.page");
+	}
+	if (UseOutputPane) {
+		SendOutput(SCI_SETSEL,0,0);
+	} else {
+		SendEditor(SCI_SETCODEPAGE, codePage);
+
+		// Zero all the style bytes
+		SendEditor(SCI_CLEARDOCUMENTSTYLE);
+
+		CurrentBuffer()->overrideExtension = "x.txt";
+		ReadProperties();
+		SetIndentSettings();
+		SendEditor(SCI_COLOURISE, 0, -1);
+		Redraw();
+
+		SendEditor(SCI_SETSEL,0,0);
+	}
+}
+
+void SciTEBase::OpenFilesFromStdin() {
+	char data[blockSize];
+	char *pNL;
+
+	/* if stdin is blocked, do not execute this method */
+	if (IsStdinBlocked())
+		return;
+
+	while (fgets(data, sizeof(data) - 1, stdin)) {
+		if ((pNL = strchr(data,'\n')) != NULL)
+			*pNL = '\0';
+		Open(data, ofQuiet);
+	}
+	if (buffers.length == 0)
+		Open("");
+}
+
+class BufferedFile {
+	FILE *fp;
+	bool readAll;
+	bool exhausted;
+	enum {bufLen = 64*1024};
+	char buffer[bufLen];
+	size_t pos;
+	size_t valid;
+public:
+	BufferedFile(FilePath fPath) {
+		fp = fPath.Open(fileRead);
+		readAll = false;
+		exhausted = fp == NULL;
+		pos = 0;
+		valid = 0;
+	}
+	~BufferedFile() {
+		if (fp) {
+			fclose(fp);
+		}
+		fp = NULL;
+	}
+	bool Exhausted() {
+		return exhausted;
+	}
+	int NextByte() {
+		if (pos >= valid) {
+			if (readAll) {
+				exhausted = true;
+			} else {
+				valid = fread(buffer, 1, bufLen, fp);
+				if (valid < bufLen) {
+					readAll = true;
+				}
+				pos = 0;
+			}
+			if (pos >= valid) {
+				return 0;
+			}
+		}
+		return buffer[pos++];
+	}
+};
+
+class FileReader {
+	BufferedFile *bf;
+	int lineNum;
+	bool lastWasCR;
+	enum {bufLen = 1000};
+	char lineToCompare[bufLen+1];
+	char lineToShow[bufLen+1];
+	bool caseSensitive;
+public:
+
+	FileReader(FilePath fPath, bool caseSensitive_) {
+		bf = new BufferedFile(fPath);
+		lineNum = 0;
+		lastWasCR = false;
+		caseSensitive = caseSensitive_;
+	}
+	~FileReader() {
+		delete bf;
+		bf = NULL;
+	}
+	char *Next() {
+		if (bf->Exhausted()) {
+			return NULL;
+		}
+		int i = 0;
+		while (!bf->Exhausted()) {
+			int ch = bf->NextByte();
+			if (i < bufLen) {
+				if (i == 0 && lastWasCR && ch == '\n') {
+					lastWasCR = false;
+					ch = 0;
+				} else if (ch == '\r' || ch == '\n') {
+					lastWasCR = ch == '\r';
+					break;
+				} else {
+					lineToShow[i++] = static_cast<char>(ch);
+				}
+			}
+		}
+		lineToShow[i] = '\0';
+		lineNum++;
+		strcpy(lineToCompare, lineToShow);
+		if (!caseSensitive) {
+			for (int j=0; j<i; j++) {
+				if (lineToCompare[j] >= 'A' && lineToCompare[j] <= 'Z') {
+					lineToCompare[j] = static_cast<char>(lineToCompare[j] - 'A' + 'a');
+				}
+			}
+		}
+		return lineToCompare;
+	}
+	int LineNumber() {
+		return lineNum;
+	}
+	const char *Original() {
+		return lineToShow;
+	}
+};
+
+static bool IsWordCharacter(int ch) {
+	return (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z')  || (ch >= '0' && ch <= '9')  || (ch == '_');
+}
+
+void SciTEBase::GrepRecursive(GrepFlags gf, FilePath baseDir, const char *searchString, const char *fileTypes) {
+	FilePathSet directories;
+	FilePathSet files;
+	baseDir.List(directories, files);
+	size_t searchLength = strlen(searchString);
+	SString os;
+	for (size_t i = 0; i < files.Length(); i ++) {
+		FilePath fPath = files.At(i);
+		if (fPath.Matches(fileTypes)) {
+			//OutputAppendStringSynchronised(i->AsFileSystem());
+			//OutputAppendStringSynchronised("\n");
+			FileReader fr(fPath, gf & grepMatchCase);
+			while (char *line=fr.Next()) {
+				char *match = strstr(line, searchString);
+				if (match) {
+					if (gf & grepWholeWord) {
+						char *lineEnd = line + strlen(line);
+						while (match) {
+							if (((match == line) || !IsWordCharacter(match[-1])) &&
+								((match+searchLength == (lineEnd)) || !IsWordCharacter(match[searchLength]))) {
+									break;
+							}
+							match = strstr(match+1, searchString);
+						}
+					}
+					if (match) {
+						os.append(fPath.AsFileSystem());
+						os.append(":");
+						SString lNumber(fr.LineNumber());
+						os.append(lNumber.c_str());
+						os.append(":");
+						os.append(fr.Original());
+						os.append("\n");
+					}
+				}
+			}
+		}
+	}
+	if (os.length()) {
+		if (gf & grepStdOut) {
+			fwrite(os.c_str(), os.length(), 1, stdout);
+		} else {
+			OutputAppendStringSynchronised(os.c_str());
+		}
+	}
+	for (size_t j = 0; j < directories.Length(); j++) {
+		FilePath fPath = directories.At(j);
+		GrepRecursive(gf, fPath, searchString, fileTypes);
+	}
+}
+
+void SciTEBase::InternalGrep(GrepFlags gf, const char *directory, const char *fileTypes, const char *search) {
+	int originalEnd = 0;
+	ElapsedTime commandTime;
+	if (!(gf & grepStdOut)) {
+		SString os;
+		os.append(">Internal search for \"");
+		os.append(search);
+		os.append("\" in \"");
+		os.append(fileTypes);
+		os.append("\"\n");
+		OutputAppendStringSynchronised(os.c_str());
+		MakeOutputVisible();
+		originalEnd = SendOutputEx(SCI_GETCURRENTPOS, 0, 0, false);
+	}
+	SString searchString(search);
+	if (!(gf & grepMatchCase)) {
+ 		searchString.lowercase();
+	}
+	GrepRecursive(gf, FilePath(directory), searchString.c_str(), fileTypes);
+	if (!(gf & grepStdOut)) {
+		SString sExitMessage(">");
+		if (timeCommands) {
+			sExitMessage += "    Time: ";
+			sExitMessage += SString(commandTime.Duration(), 3);
+		}
+		sExitMessage += "\n";
+		OutputAppendStringSynchronised(sExitMessage.c_str());
+		if (props.GetInt("output.scroll", 1) == 1 && returnOutputToCommand)
+			SendOutputEx(SCI_GOTOPOS, originalEnd, 0, false);
+	}
+}
