@@ -341,7 +341,7 @@ void SciTEWin::SaveAsTEX() {
 
 void SciTEWin::SaveAsXML() {
 	FilePath path = ChooseSaveName(filePath.Directory(), "Export File As XML",
-	                              "TeX (.xml)\0*.xml\0", ".xml");
+	                              "XML (.xml)\0*.xml\0", ".xml");
 	if (path.IsSet()) {
 		SaveToXML(path);
 	}
@@ -391,6 +391,7 @@ void SciTEWin::SaveSessionDialog() {
 void SciTEWin::Print(
     bool showDialog) {	///< false if must print silently (using default settings).
 
+	RemoveFindMarks();
 	PRINTDLG pdlg = {
 	                    sizeof(PRINTDLG), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 	                };
